/*
 * StateEliminationUnbounded.cpp
 *
 *  Created on: 9. 2. 2014
 *      Author: Tomas Pecka
 */

#include "StateEliminationUnbounded.h"

#include <regexp/unbounded/UnboundedRegExp.h>
#include <exception/AlibException.h>

#include "../../../regexp/RegExpOptimize.h"

namespace conversions
{

namespace fa2re
{

regexp::RegExp StateEliminationUnbounded::convert(const automaton::Automaton& automaton)
{
	regexp::RegExp* out = NULL;
	automaton.getData().Accept((void*) &out, StateEliminationUnbounded::STATE_ELIMINATION_UNBOUNDED);
	regexp::RegExp res = std::move(*out);
	delete out;
	return res;
}

template<class T>
regexp::UnboundedRegExp StateEliminationUnbounded::convert(const T& automaton)
{
    if(automaton.getFinalStates().size() == 0)
        return regexp::UnboundedRegExp(regexp::UnboundedRegExpEmpty());

    // steps 1 + 2
    automaton::ExtendedNFA extendedAutomaton = constructExtendedNFA(automaton);
    extendExtendedNFA(extendedAutomaton);

    // step 3 - Exterminate!
    // select all states that are neither final nor initial
    std::set<automaton::State> statesToEliminate = extendedAutomaton.getStates();
    statesToEliminate.erase(*extendedAutomaton.getInitialStates().begin());
    statesToEliminate.erase(*extendedAutomaton.getFinalStates().begin());

    for(const auto& state : statesToEliminate)
        extendedAutomaton = eliminateState(extendedAutomaton, state);

    // step 4
    regexp::RegExpOptimize opt;

    regexp::UnboundedRegExpConcatenation concat;
    concat.appendElement(std::move(*transition(extendedAutomaton, *extendedAutomaton.getInitialStates().begin(), *extendedAutomaton.getFinalStates().begin())));
    concat.appendElement(regexp::UnboundedRegExpIteration(std::move(*transition(extendedAutomaton, *extendedAutomaton.getFinalStates().begin(), *extendedAutomaton.getFinalStates().begin()))));
    return opt.optimize(regexp::UnboundedRegExp(concat));
}


void StateEliminationUnbounded::Visit(void*, const automaton::UnknownAutomaton&) const
{
    throw exception::AlibException("Unsupported automaton type UnknownAutomaton");
}

void StateEliminationUnbounded::Visit(void* data, const automaton::EpsilonNFA& automaton) const
{
    regexp::RegExp* & out = *((regexp::RegExp**) data);
    out = new regexp::RegExp(convert(automaton));
}

void StateEliminationUnbounded::Visit(void* data, const automaton::NFA& automaton) const
{
    regexp::RegExp* & out = *((regexp::RegExp**) data);
    out = new regexp::RegExp(convert(automaton));
}

void StateEliminationUnbounded::Visit(void* data, const automaton::DFA& automaton) const
{
    regexp::RegExp* & out = *((regexp::RegExp**) data);
    out = new regexp::RegExp(convert(automaton));
}

void StateEliminationUnbounded::Visit(void* data, const automaton::ExtendedNFA& automaton) const
{
    regexp::RegExp* & out = *((regexp::RegExp**) data);
    out = new regexp::RegExp(convert(automaton));
}

void StateEliminationUnbounded::Visit(void*, const automaton::CompactNFA& ) const
{
    throw exception::AlibException("Unsupported automaton type CompactNFA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::DPDA&) const
{
    throw exception::AlibException("Unsupported automaton type DPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::SinglePopDPDA&) const
{
    throw exception::AlibException("Unsupported automaton type SinglePopDPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::InputDrivenNPDA&) const
{
    throw exception::AlibException("Unsupported automaton type NPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::VisiblyPushdownDPDA&) const
{
    throw exception::AlibException("Unsupported automaton type VisiblyPushdownDPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::VisiblyPushdownNPDA&) const
{
    throw exception::AlibException("Unsupported automaton type VisiblyPushdownNPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::RealTimeHeightDeterministicDPDA&) const
{
    throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicDPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::RealTimeHeightDeterministicNPDA&) const
{
    throw exception::AlibException("Unsupported automaton type RealTimeHeightDeterministicNPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::NPDA&) const
{
    throw exception::AlibException("Unsupported automaton type NPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::SinglePopNPDA&) const
{
    throw exception::AlibException("Unsupported automaton type SinglePopNPDA");
}

void StateEliminationUnbounded::Visit(void*, const automaton::OneTapeDTM&) const
{
    throw exception::AlibException("Unsupported automaton type OneTapeDTM");
}


automaton::ExtendedNFA StateEliminationUnbounded::eliminateState(const automaton::ExtendedNFA& extendedAutomaton, const automaton::State& q)
{
    automaton::ExtendedNFA newAutomaton;
    newAutomaton.setStates(extendedAutomaton.getStates());
    newAutomaton.removeState(q); // preserve all states but q (the one to eliminate)
    newAutomaton.setInputSymbols(extendedAutomaton.getInputAlphabet());
    newAutomaton.setInitialStates(extendedAutomaton.getInitialStates()); // sure that q is neither initial nor final (follows from step 2 - extending ExtendedNFA)
    newAutomaton.setFinalStates(extendedAutomaton.getFinalStates());

    regexp::RegExpOptimize opt;

    for(const auto& p: newAutomaton.getStates())
    {
        for(const auto& r : newAutomaton.getStates())
        {
            regexp::UnboundedRegExpConcatenation concat;
            concat.appendElement(std::move(*transition(extendedAutomaton, p, q)));
            concat.appendElement(regexp::UnboundedRegExpIteration(std::move(*transition(extendedAutomaton, q, q))));
            concat.appendElement(std::move(*transition(extendedAutomaton, q, r)));

            regexp::UnboundedRegExpAlternation alt;
            alt.appendElement(std::move(*transition(extendedAutomaton, p, r)));
            alt.appendElement(concat);

            newAutomaton.addTransition(p, regexp::RegExp{regexp::UnboundedRegExp(opt.optimize(regexp::UnboundedRegExp(alt)))}, r);
        }
    }

    return newAutomaton;
}

const regexp::UnboundedRegExpElement* StateEliminationUnbounded::transition(const automaton::ExtendedNFA& automaton, const automaton::State& from, const automaton::State& to)
{
    regexp::RegExpOptimize opt;
    regexp::UnboundedRegExpAlternation ret;

    for(const auto& transition: automaton.getTransitionsFromState(from))
        if(transition.second.count(to) > 0)
            ret.appendElement(static_cast<const regexp::UnboundedRegExp&>(transition.first.second.getData()).getRegExp());

    return opt.optimize(regexp::UnboundedRegExp(ret)).getRegExp().clone();
}

template<>
automaton::ExtendedNFA StateEliminationUnbounded::constructExtendedNFA(const automaton::NFA& automaton)
{
    automaton::ExtendedNFA extendedAutomaton;
    extendedAutomaton.setStates(automaton.getStates());
    extendedAutomaton.setInputSymbols(automaton.getInputAlphabet());
    extendedAutomaton.setInitialStates(automaton.getInitialStates());
    extendedAutomaton.setFinalStates(automaton.getFinalStates());

    for(const auto& transition : automaton.getTransitions()) // pair<State, symb/eps> -> set<State>
        for(const auto& to : transition.second)
            extendedAutomaton.addTransition(transition.first.first, regexp::RegExp{regexp::UnboundedRegExp(regexp::UnboundedRegExpSymbol(transition.first.second))}, to);

    return extendedAutomaton;
}

template<>
automaton::ExtendedNFA StateEliminationUnbounded::constructExtendedNFA(const automaton::DFA& automaton)
{
    automaton::ExtendedNFA extendedAutomaton;
    extendedAutomaton.setStates(automaton.getStates());
    extendedAutomaton.setInputSymbols(automaton.getInputAlphabet());
    extendedAutomaton.setInitialStates(std::set<automaton::State>{automaton.getInitialState()});
    extendedAutomaton.setFinalStates(automaton.getFinalStates());

    for(const auto& transition : automaton.getTransitions()) // pair<State, symb/eps> -> set<State>
        extendedAutomaton.addTransition(transition.first.first, regexp::RegExp{regexp::UnboundedRegExp(regexp::UnboundedRegExpSymbol(transition.first.second))}, transition.second);

    return extendedAutomaton;
}

template<>
automaton::ExtendedNFA StateEliminationUnbounded::constructExtendedNFA(const automaton::EpsilonNFA& automaton)
{
    automaton::ExtendedNFA extendedAutomaton;
    extendedAutomaton.setStates(automaton.getStates());
    extendedAutomaton.setInputSymbols(automaton.getInputAlphabet());
    extendedAutomaton.setInitialStates(automaton.getInitialStates());
    extendedAutomaton.setFinalStates(automaton.getFinalStates());

    for(const auto& transition : automaton.getSymbolTransitions()) // pair<State, symb/eps> -> set<State>
        for(const auto& to : transition.second)
            extendedAutomaton.addTransition(transition.first.first, regexp::RegExp{regexp::UnboundedRegExp(regexp::UnboundedRegExpSymbol(transition.first.second))}, to);

    for(const auto& transition : automaton.getEpsilonTransitions())
        for(const auto& to : transition.second)
            extendedAutomaton.addTransition(transition.first, regexp::RegExp{regexp::UnboundedRegExp(regexp::UnboundedRegExpEpsilon())}, to);

    return extendedAutomaton;
}

template<>
automaton::ExtendedNFA StateEliminationUnbounded::constructExtendedNFA(const automaton::ExtendedNFA& automaton)
{
    return automaton;
}

void StateEliminationUnbounded::extendExtendedNFA(automaton::ExtendedNFA& automaton)
{
    const automaton::State& initState = *automaton.getInitialStates().begin();
    if(automaton.getFinalStates().count(initState) > 0 || automaton.getTransitionsToState(initState).size() > 0 )
    {
        const automaton::State q0 = automaton::createUniqueState(initState, automaton.getStates());
        automaton.addState(q0);

        regexp::UnboundedRegExp regexp;
        regexp.setRegExp(regexp::UnboundedRegExpEpsilon());
        automaton.addTransition(q0, regexp::RegExp{regexp}, initState);

        const std::set<automaton::State> automatonInitialStates = automaton.getInitialStates();
        for(const auto& state: automatonInitialStates)
            automaton.removeInitialState(state);
        automaton.addInitialState(q0);
    }

    if(automaton.getFinalStates().size() > 1)
    {
        const automaton::State f = automaton::createUniqueState(automaton::State("f"), automaton.getStates());
        automaton.addState(f);

        const std::set<automaton::State> automatonFinalStates = automaton.getFinalStates();
        for(const auto &state : automatonFinalStates)
        {
            regexp::UnboundedRegExp regexp;
            regexp.setRegExp(regexp::UnboundedRegExpEpsilon());
            automaton.addTransition(state, regexp::RegExp{regexp}, f);

            automaton.removeFinalState(state);
        }

        automaton.addFinalState(f);
    }
}

const StateEliminationUnbounded StateEliminationUnbounded::STATE_ELIMINATION_UNBOUNDED;

} /* namespace fa2re */

} /* namespace conversions */
