/*
 * GlushkovNFA.cpp
 *
 *  Created on: 11. 1. 2014
 *	  Author: Tomas Pecka
 */

#include "GlushkovNFA.h"

#include "label/Label.h"
#include "label/LabelPairLabel.h"

#include "../../regexp/GlushkovTraversal.h"
#include "../../regexp/GlushkovPair.h"
#include "../../regexp/GlushkovSymbol.h"

#include "../../regexp/RegExpEpsilon.h"

namespace conversions{

namespace re2fa {

automaton::NFA GlushkovNFA::convert(const regexp::RegExp& regexp) {
	automaton::NFA* out = nullptr;
	regexp.getData().Accept((void*) &out, GlushkovNFA::GLUSHKOV_NFA);
	automaton::NFA res = std::move(*out);
	delete out;
	return res;
}

automaton::NFA GlushkovNFA::convert(const regexp::UnboundedRegExp& regexp)
{
	automaton::NFA automaton;

	// step 1
	for( auto const& symbol : regexp.getAlphabet( ) )
		automaton.addInputSymbol( symbol );

	// steps 2, 3, 4
    std::set<regexp::GlushkovPair> pairs;
    const std::set<regexp::GlushkovSymbol> first = regexp::GlushkovTraversal::first(regexp);
    const std::set<regexp::GlushkovSymbol> last = regexp::GlushkovTraversal::last(regexp);
	for( auto const& x : regexp::GlushkovTraversal::getSymbols( regexp ) )
		for( auto const& f : regexp::GlushkovTraversal::follow( regexp, x ) ) {
			pairs.insert( regexp::GlushkovPair( x, f ) );
		}
	// \e in q0 check is in step 7

	// step 5
    std::map<regexp::GlushkovSymbol, automaton::State> stateMap;

	automaton::State q0( label::Label( label::LabelPairLabel( std::make_pair( label::labelFrom( 'q' ), label::labelFrom( 0 ) ) ) ) );
	automaton.addState( q0 );
	automaton.addInitialState( q0 );
	for( auto const& symbol : regexp::GlushkovTraversal::getSymbols( regexp ) ) {
		automaton::State q( label::Label( label::LabelPairLabel( std::make_pair( label::labelFrom( symbol.getInputSymbol( ) ), label::labelFrom( symbol.getId( ) ) ) ) ) );

		stateMap.insert( std::make_pair( symbol, q ) );
		automaton.addState( q );
	}

	// step 6
	for( auto const& symbol : first ) {
		const automaton::State & q = stateMap.find( symbol )->second;

		automaton.addTransition( q0, symbol.getInputSymbol( ), q );
	}

	for( auto const& pair : pairs ) {
		const automaton::State & p = stateMap.find( pair.getFirst( ) )->second;
		const automaton::State & q = stateMap.find( pair.getSecond( ) )->second;

		automaton.addTransition( p, pair.getSecond( ).getInputSymbol( ), q );
	}

	// step 7
	for( auto const& symbol : last ) {
		const automaton::State & q = stateMap.find( symbol )->second;
		automaton.addFinalState( q );
	}

	if(regexp::RegExpEpsilon::languageContainsEpsilon(regexp))
		automaton.addFinalState( q0 );

	return automaton;
}

void GlushkovNFA::Visit(void* userData, const regexp::FormalRegExp& regexp) const
{
	/*
	automaton::NFA* & out = *((automaton::NFA**) userData);
	out = new automaton::NFA(this->convert(regexp));
	*/
	throw exception::AlibException("Glushkov: Converting FormalRegExp NYI"); // TODO
}

void GlushkovNFA::Visit(void* userData, const regexp::UnboundedRegExp& regexp) const
{
	automaton::NFA* & out = *((automaton::NFA**) userData);
	out = new automaton::NFA(this->convert(regexp));
}

const GlushkovNFA GlushkovNFA::GLUSHKOV_NFA;

} /* namespace fa2re */

} /* namespace conversions */
