/*
 * Thompson.h
 *
 *  Created on: 11. 1. 2014
 *	  Author: Tomas Pecka
 */

#ifndef THOMPSON_H_
#define THOMPSON_H_

#include <regexp/RegExp.h>
#include <regexp/formal/FormalRegExpElements.h>
#include <regexp/unbounded/UnboundedRegExpElements.h>
#include <automaton/Automaton.h>
#include <automaton/FSM/EpsilonNFA.h>

namespace conversions
{

namespace re2fa
{

/**
 * Converts regular expression to finite automaton using Thompson's Construction Algorithm (TCA).
 * Sources:
 *  Hopcroft, section 3.2.3
 *  http://www.eecis.udel.edu/~cavazos/cisc672/lectures/Lecture-04.pdf
 *  http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.21.7450&rep=rep1&type=ps
 *  Melichar 2.112
 */
class Thompson : public regexp::VisitableRegExpBase::const_visitor_type, regexp::FormalRegExpElement::const_visitor_type, regexp::UnboundedRegExpElement::const_visitor_type
{
public:
	/**
	 * Performs conversion.
	 * @param regexp regexp to convert
	 * @return nondeterministic finite automaton with epsilon transitions accepting language described by the regexp
	 */
	static automaton::Automaton convert(const regexp::RegExp& regexp);

	template<class T>
	static automaton::EpsilonNFA convert(const T& regexp);

private:
	 void Visit(void*, const regexp::UnboundedRegExp& regexp) const;
	 void Visit(void*, const regexp::FormalRegExp& regexp) const;

	 void Visit(void*, const regexp::UnboundedRegExpAlternation& alternation) const;
	 void Visit(void*, const regexp::UnboundedRegExpConcatenation& concatenation) const;
	 void Visit(void*, const regexp::UnboundedRegExpIteration& iteration) const;
	 void Visit(void*, const regexp::UnboundedRegExpSymbol& symbol) const;
	 void Visit(void*, const regexp::UnboundedRegExpEpsilon& epsilon) const;
	 void Visit(void*, const regexp::UnboundedRegExpEmpty& empty) const;

	 void Visit(void*, const regexp::FormalRegExpAlternation& alternation) const;
	 void Visit(void*, const regexp::FormalRegExpConcatenation& concatenation) const;
	 void Visit(void*, const regexp::FormalRegExpIteration& iteration) const;
	 void Visit(void*, const regexp::FormalRegExpSymbol& symbol) const;
	 void Visit(void*, const regexp::FormalRegExpEpsilon& epsilon) const;
	 void Visit(void*, const regexp::FormalRegExpEmpty& empty) const;

	 static const Thompson THOMPSON;
};

} /* namespace re2fa */

} /* namespace conversions */

#endif /* THOMPSON_H_ */
