/*
 * RGtoFA.cpp
 *
 * Created on: 1. 11. 2013
 * Author: Tomas Pecka
 */
#include "RGtoFA.h"

#include <alphabet/LabeledSymbol.h>
#include <exception/AlibException.h>

namespace conversions
{

namespace rg2fa
{

automaton::Automaton RGtoFA::convert(const grammar::Grammar& grammar)
{
     automaton::Automaton* out = NULL;
     grammar.getData().Accept((void*) &out, RGtoFA::RG_TO_FA_CONVERSION);
     automaton::Automaton res = std::move(*out);
     delete out;
     return res;
}

automaton::NFA RGtoFA::convert(const grammar::LeftRG& grammar)
{
    std::map<alphabet::Symbol, automaton::State> stateMap;

    // step 1
    automaton::NFA automaton;
    automaton.setInputSymbols(grammar.getTerminalAlphabet());

    // step 2
    for(const auto& symbol : grammar.getNonterminalAlphabet())
    {
        automaton::State state(static_cast<const alphabet::LabeledSymbol&>(symbol.getData()).getLabel());
        automaton.addState(state);
        stateMap.insert(std::make_pair(symbol, state));
    }
    const automaton::State q0 = automaton::createUniqueState(automaton::State("q0"), automaton.getStates());
    automaton.addState(q0);

    // step 3
    for(const auto& rule : grammar.getRules())
    {
        const alphabet::Symbol& lhs = rule.first;
        for(const auto& ruleRHS : rule.second)
        {
            if(ruleRHS.is<std::pair<alphabet::Symbol, alphabet::Symbol>>()) // if B->Ca => \delta(C,a)=B
            {
                const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = ruleRHS.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
                automaton.addTransition(stateMap.find(rhs.first)->second, rhs.second, stateMap.find(lhs)->second);
            }
            else // if B->a => \delta(StartState,a)=B
            {
                const alphabet::Symbol& rhs = ruleRHS.get<alphabet::Symbol>();
                automaton.addTransition(q0, rhs, stateMap.find(lhs)->second);
            }
        }
    }

    // step 4
    automaton.addInitialState(q0);

    // step 5
    automaton.addFinalState(stateMap.find(grammar.getInitialSymbol())->second);
    if(grammar.getGeneratesEpsilon())
        automaton.addFinalState(q0);

    return automaton;
}

automaton::NFA RGtoFA::convert(const grammar::RightRG& grammar)
{
    std::map<alphabet::Symbol, automaton::State> stateMap;

    // step 1
    automaton::NFA automaton;
    automaton.setInputSymbols(grammar.getTerminalAlphabet());

    // step2
    for(const auto& symbol : grammar.getNonterminalAlphabet())
    {
        automaton::State state(static_cast<const alphabet::LabeledSymbol&>(symbol.getData()).getLabel());
        automaton.addState(state);
        stateMap.insert(std::make_pair(symbol, state));
    }
    const automaton::State AState = automaton::createUniqueState(automaton::State("A"), automaton.getStates());
    automaton.addState(AState);

    // step 3
    for(const auto& rule : grammar.getRules())
    {
        const alphabet::Symbol& lhs = rule.first;
        for(const auto& ruleRHS : rule.second)
        {
            if(ruleRHS.is<std::pair<alphabet::Symbol, alphabet::Symbol>>()) // if B->aC => \delta(B,a)=C
            {
                const std::pair<alphabet::Symbol, alphabet::Symbol>& rhs = ruleRHS.get<std::pair<alphabet::Symbol, alphabet::Symbol>>();
                automaton.addTransition(stateMap.find(lhs)->second, rhs.first, stateMap.find(rhs.second)->second);
            }
            else // if B->a => \delta(B,a)=AState
            {
                const alphabet::Symbol& rhs = ruleRHS.get<alphabet::Symbol>();
                automaton.addTransition(stateMap.find(lhs)->second, rhs, AState);
            }
        }
    }

    // step 4
    automaton.addInitialState(stateMap.find(grammar.getInitialSymbol())->second);

    // step 5
    automaton.addFinalState(AState);
    if(grammar.getGeneratesEpsilon())
        for(const auto& state : automaton.getInitialStates())
            automaton.addFinalState(state);

    return automaton;
}

void RGtoFA::Visit(void*, const grammar::UnknownGrammar&) const
{
    throw exception::AlibException("Unsupported grammar type UnknownGrammar");
}

void RGtoFA::Visit(void*, const grammar::ContextPreservingUnrestrictedGrammar&) const
{
    throw exception::AlibException("Unsupported grammar type ContextPreservingUnrestrictedGrammar");
}

void RGtoFA::Visit(void*, const grammar::UnrestrictedGrammar&) const
{
    throw exception::AlibException("Unsupported grammar type UnrestrictedGrammar");
}

void RGtoFA::Visit(void*, const grammar::CSG&) const
{
    throw exception::AlibException("Unsupported grammar type CSG");
}

void RGtoFA::Visit(void*, const grammar::NonContractingGrammar&) const
{
    throw exception::AlibException("Unsupported grammar type NonContractingGrammar");
}

void RGtoFA::Visit(void*, const grammar::CNF&) const
{
    throw exception::AlibException("Unsupported grammar type CNF");
}

void RGtoFA::Visit(void*, const grammar::CFG&) const
{
    throw exception::AlibException("Unsupported grammar type CFG");
}

void RGtoFA::Visit(void*, const grammar::EpsilonFreeCFG&) const
{
    throw exception::AlibException("Unsupported grammar type EpsilonFreeCFG");
}

void RGtoFA::Visit(void*, const grammar::GNF&) const
{
    throw exception::AlibException("Unsupported grammar type GNF");
}

void RGtoFA::Visit(void*, const grammar::LG&) const
{
    throw exception::AlibException("Unsupported grammar type LG");
}

void RGtoFA::Visit(void* userData, const grammar::RightRG& grammar) const
{
    automaton::Automaton* & out = *((automaton::Automaton**) userData);
    out = new automaton::Automaton(this->convert(grammar));
}

void RGtoFA::Visit(void* userData, const grammar::LeftRG& grammar) const
{
    automaton::Automaton* & out = *((automaton::Automaton**) userData);
    out = new automaton::Automaton(this->convert(grammar));
}

void RGtoFA::Visit(void*, const grammar::RightLG&) const
{
    throw exception::AlibException("Unsupported grammar type RightLG");
}

void RGtoFA::Visit(void*, const grammar::LeftLG&) const
{
    throw exception::AlibException("Unsupported grammar type LeftLG");
}

const RGtoFA RGtoFA::RG_TO_FA_CONVERSION;

} /* namespace rg2fa */

} /* namespace conversions */
