/*
 * SinglePopDPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef SINGLE_POP_DPDA_H_
#define SINGLE_POP_DPDA_H_

#include <set>
#include <map>
#include <vector>
#include "../../std/variant.hpp"
#include "../AutomatonBase.h"
#include "../common/SingleInitialState.h"
#include "../common/SingleInitialSymbolPushdownStoreAlphabet.h"
#include "../common/InputAlphabet.h"
#include "../../alphabet/Symbol.h"
#include "../../string/Epsilon.h"

namespace automaton {

/**
 * Push Down Automaton
 *
 * $|\delta (q, a, \gamma)| \leq 1$, $\forall q, a, \gamma, q \in Q, a \in (T \cup \{\varepsilon\}), \gamma \in G^*.$
 * if $\delta (q, a, r) \neq \emptyset$, $\delta (q, a, s) \neq \emptyset$ then $r \neq s$.
 * if $\delta(q, a, r) \neq \emptyset$, $\delta (q, \varepsilon, s) \neq \emptyset$, then $r \neq s$.
 */
class SinglePopDPDA: public std::acceptor<SinglePopDPDA, VisitableAutomatonBase, std::acceptor<SinglePopDPDA, alib::VisitableObjectBase, AutomatonBase> >, public SingleInitialSymbolPushdownStoreAlphabet, public SingleInitialState, public InputAlphabet {
protected:
	std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::pair<State, std::vector<alphabet::Symbol> > > transitions;
public:
	explicit SinglePopDPDA(const State& initialState, const alphabet::Symbol& initialPushdownSymbol);

	virtual AutomatonBase* clone() const;
	
	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol(const alphabet::Symbol& symbol);

	/**
	 * @copydoc Automaton::removeStackSymbol(const Symbol&)
	 */
	virtual bool removeStackSymbol(const alphabet::Symbol& symbol);

	/**
	 * Adds transition to the SinglePopDPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the SinglePopDPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);

	bool addTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);
	
	bool addTransition(const State& from, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);

	/**
	 * Removes the transition from the SinglePopDPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the SinglePopDPDA
	 */
	bool removeTransition(const State& from, const std::variant<string::Epsilon, alphabet::Symbol>& input, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);

	bool removeTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);
	
	bool removeTransition(const State& from, const alphabet::Symbol& pop, const State& to, const std::vector<alphabet::Symbol>& push);

	/**
	 * @return SinglePopDPDA transitions
	 */
	const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, std::pair<State, std::vector<alphabet::Symbol> > >& getTransitions() const;

	virtual bool operator<(const alib::ObjectBase& other) const;
	virtual bool operator==(const alib::ObjectBase& other) const;
	virtual bool operator>(const alib::ObjectBase& other) const;

	virtual bool operator==(const SinglePopDPDA& other) const;
	virtual bool operator<(const SinglePopDPDA& other) const;

	virtual void operator>>(std::ostream& os) const;

	virtual operator std::string() const;

	virtual int selfTypeId() const {
		return typeId(*this);
	}
};

} /* namespace automaton */

#endif /* SINGLE_POP_DPDA_H_ */
