/*
 * LabelFromXMLParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LabelFromXMLParser.h"
#include "../sax/ParserException.h"
#include "../Api.hpp"
#include <algorithm>

namespace label {

Label LabelFromXMLParser::parseLabel(std::list<sax::Token>& input) const {
	return parseLabel(input, std::set<FEATURES>({FEATURES::PRIMITIVE, FEATURES::HEXAVIGESIMAL, FEATURES::OBJECT, FEATURES::LABEL_SET, FEATURES::LABEL_PAIR}));
}

Label LabelFromXMLParser::parseLabel(std::list<sax::Token>& input, const std::set<FEATURES>& features) const {
	if(isToken(input, sax::Token::TokenType::START_ELEMENT, "PrimitiveLabel")) {
		if(!features.count(FEATURES::PRIMITIVE)) throw exception::AlibException();
		return Label(parsePrimitiveLabel(input));
	} else if(isToken(input, sax::Token::TokenType::START_ELEMENT, "HexavigesimalLabel")) {
		if(!features.count(FEATURES::HEXAVIGESIMAL)) throw exception::AlibException();
		return Label(parseHexavigesimalLabel(input));
	} else if(isToken(input, sax::Token::TokenType::START_ELEMENT, "ObjectLabel")) {
		if(!features.count(FEATURES::OBJECT)) throw exception::AlibException();
		return Label(parseObjectLabel(input));
	} else if(isToken(input, sax::Token::TokenType::START_ELEMENT, "LabelSetLabel")) {
		if(!features.count(FEATURES::LABEL_SET)) throw exception::AlibException();
		return Label(parseLabelSetLabel(input));
	} else if(isToken(input, sax::Token::TokenType::START_ELEMENT, "LabelPairLabel")) {
		if(!features.count(FEATURES::LABEL_PAIR)) throw exception::AlibException();
		return Label(parseLabelPairLabel(input));
	} else {
		throw sax::ParserException(sax::Token("PrimitiveLabel, ObjectLabel, LabelSetLabel, LabelPairLabel", sax::Token::TokenType::START_ELEMENT), input.front());
	}
}

bool LabelFromXMLParser::first(std::list<sax::Token>& input) const {
	if(isToken(input, sax::Token::TokenType::START_ELEMENT, "PrimitiveLabel") || isToken(input, sax::Token::TokenType::START_ELEMENT, "ObjectLabel") || isToken(input, sax::Token::TokenType::START_ELEMENT, "LabelSetLabel") || isToken(input, sax::Token::TokenType::START_ELEMENT, "LabelPairLabel")) {
		return true;
	} else {
		return false;
	}
}

PrimitiveLabel LabelFromXMLParser::parsePrimitiveLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "PrimitiveLabel");
	PrimitiveLabel data(alib::api<primitive::Primitive>::parse(input));
	popToken(input, sax::Token::TokenType::END_ELEMENT, "PrimitiveLabel");
	return data;
}

HexavigesimalLabel LabelFromXMLParser::parseHexavigesimalLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "HexavigesimalLabel");
	HexavigesimalLabel data(std::stoi(popTokenData(input, sax::Token::TokenType::CHARACTER)));
	popToken(input, sax::Token::TokenType::END_ELEMENT, "HexavigesimalLabel");
	return data;
}

ObjectLabel LabelFromXMLParser::parseObjectLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "ObjectLabel");
	ObjectLabel data(alib::api<alib::Object>::parse(input));
	popToken(input, sax::Token::TokenType::END_ELEMENT, "ObjectLabel");
	return data;
}

LabelSetLabel LabelFromXMLParser::parseLabelSetLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "LabelSetLabel");
	std::set<Label> labels;
	while(isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		labels.insert(parseLabel(input));
	}
	LabelSetLabel data(labels);
	popToken(input, sax::Token::TokenType::END_ELEMENT, "LabelSetLabel");
	return data;
}

LabelPairLabel LabelFromXMLParser::parseLabelPairLabel(std::list<sax::Token>& input) const {
	popToken(input, sax::Token::TokenType::START_ELEMENT, "LabelPairLabel");
	
	Label firstLabel = parseLabel(input);
	Label secondLabel = parseLabel(input);

	LabelPairLabel data(std::make_pair(firstLabel, secondLabel));
	popToken(input, sax::Token::TokenType::END_ELEMENT, "LabelPairLabel");
	return data;
}

} /* namespace label */
