/*
 * RegExpFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "RegExpFromStringParser.h"
#include "../sax/ParserException.h"
#include "../exception/AlibException.h"

namespace regexp {

RegExpFromStringParser::RegExpFromStringParser(std::istream& input) : m_RegexpLexer(input), m_SymbolParser(input) {

}

RegExp RegExpFromStringParser::parse() {
	return parse(std::set<FEATURES>({FEATURES::FORMAL, FEATURES::UNBOUNDED}));
}

RegExp RegExpFromStringParser::parse(const std::set<FEATURES>& features) {
	UnboundedRegExpElement* element = this->alternation();
	UnboundedRegExp regexp(std::move(*element));
	delete element;
	if(features.count(FEATURES::UNBOUNDED)) return RegExp{regexp};

	if(features.count(FEATURES::FORMAL)) return RegExp{FormalRegExp{regexp}};

	throw exception::AlibException();
}

bool RegExpFromStringParser::first() {
	RegExpFromStringLexer::Token token = m_RegexpLexer.next().token();
	if(token.type == RegExpFromStringLexer::TokenType::EMPTY || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::LPAR) {
		return true;
	} else {
		return token.type == RegExpFromStringLexer::TokenType::ERROR && m_SymbolParser.first();
	}
}

bool RegExpFromStringParser::next() {
	RegExpFromStringLexer::Token token = m_RegexpLexer.next().token();
	if(token.type == RegExpFromStringLexer::TokenType::EMPTY || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::STAR || token.type == RegExpFromStringLexer::TokenType::RPAR || token.type == RegExpFromStringLexer::TokenType::PLUS) {
		return true;
	} else {
		return false;
	}
}

bool RegExpFromStringParser::last() {
	RegExpFromStringLexer::Token token = m_RegexpLexer.next().token();
	if(token.type == RegExpFromStringLexer::TokenType::TEOF) {
		return true;
	} else {
		return false;
	}
}

UnboundedRegExpElement* RegExpFromStringParser::alternation() {
	return this->alternationCont(this->concatenation());
}

UnboundedRegExpElement* RegExpFromStringParser::alternationCont(UnboundedRegExpElement* left) {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::PLUS) {
		next() || m_SymbolParser.first();

		try {
			UnboundedRegExpElement* right = this->concatenation();
			UnboundedRegExpAlternation* res = new UnboundedRegExpAlternation();
			res->appendElement(std::move(*left));
			res->appendElement(std::move(*right));
			delete right;
			delete left;
			left = NULL;

			return this->alternationContCont(res);
		} catch (...) {
			delete left;
			throw;
		}
	} else {
		return left;
	}
}

UnboundedRegExpElement* RegExpFromStringParser::alternationContCont(UnboundedRegExpAlternation* res) {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::PLUS) {
		next() || m_SymbolParser.first();

		try {
			UnboundedRegExpElement* next = this->concatenation();
			res->appendElement(std::move(*next));
			delete next;

			return this->alternationContCont(res);
		} catch (...) {
			delete res;
			throw;
		}
	} else {
		return res;
	}
}


UnboundedRegExpElement* RegExpFromStringParser::concatenation() {
	return this->concatenationCont(this->factor());
}

UnboundedRegExpElement* RegExpFromStringParser::concatenationCont(UnboundedRegExpElement* left) {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::ERROR || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::EMPTY) {

		try {
			UnboundedRegExpElement* right = this->factor();
			UnboundedRegExpConcatenation* res = new UnboundedRegExpConcatenation();
			res->appendElement(std::move(*left));
			res->appendElement(std::move(*right));
			delete right;
			delete left;
			left = NULL;

			return this->concatenationContCont(res);
		} catch (...) {
			delete left;
			throw;
		}
	} else {
		return left;
	}
	throw exception::AlibException("Impossible");
}

UnboundedRegExpElement* RegExpFromStringParser::concatenationContCont(UnboundedRegExpConcatenation* res) {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::ERROR || token.type == RegExpFromStringLexer::TokenType::LPAR || token.type == RegExpFromStringLexer::TokenType::EPS || token.type == RegExpFromStringLexer::TokenType::EMPTY) {

		try {
			UnboundedRegExpElement* next = this->factor();
			res->appendElement(std::move(*next));
			delete next;

			return this->concatenationContCont(res);
		} catch (...) {
			delete res;
			throw;
		}
	} else {
		return res;
	}
	throw exception::AlibException("Imúpssible");
}

UnboundedRegExpElement* RegExpFromStringParser::factor() {
	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::LPAR) {
		next() || m_SymbolParser.first();
	
		UnboundedRegExpElement* base = this->alternation();
		token = m_RegexpLexer.token();
		if(token.type != RegExpFromStringLexer::TokenType::RPAR) throw exception::AlibException();
		return this->star(base);
	} else if(token.type == RegExpFromStringLexer::TokenType::EPS) {
		return this->star(new UnboundedRegExpEpsilon());
	} else if(token.type == RegExpFromStringLexer::TokenType::EMPTY) {
		return this->star(new UnboundedRegExpEmpty());
	} else if(token.type == RegExpFromStringLexer::TokenType::ERROR) {
		UnboundedRegExpSymbol* res = new UnboundedRegExpSymbol(m_SymbolParser.parse());
		return this->star(res);
	} else {
		throw exception::AlibException("Unrecognised token at factor rule");
	}
}

UnboundedRegExpElement* RegExpFromStringParser::star(UnboundedRegExpElement* elem) {
	next() || m_SymbolParser.first();

	RegExpFromStringLexer::Token token = m_RegexpLexer.token();
	if(token.type == RegExpFromStringLexer::TokenType::STAR) {
		UnboundedRegExpIteration* iter = new UnboundedRegExpIteration(std::move(*elem));
		delete elem;
		return this->star(iter);
	} else {
		return elem;
	}
}

}
