/*
 * RegExpToStringComposer.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include <sstream>
#include <algorithm>
#include "RegExpToStringComposer.h"
#include "../alphabet/SymbolToStringComposer.h"

namespace regexp {

void RegExpToStringComposer::Visit(void* userData, const UnboundedRegExp& regexp) {
	regexp.getRegExp().Accept(userData, *this);
}

void RegExpToStringComposer::Visit(void* userData, const FormalRegExp& regexp) {
	regexp.getRegExp().Accept(userData, *this);
}

void RegExpToStringComposer::Visit(void* userData, const UnboundedRegExpAlternation& alternation) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);

	if(alternation.getElements().size() == 0) {
		out.second << "#0";
	} else if(alternation.getElements().size() == 1) {
		const auto& element = alternation.getElements()[0];
		const UnboundedRegExpElement& object = static_cast<const UnboundedRegExpElement&>(*element);
		object.Accept(userData, *this);
	} else {
		Priority outerPriorityMinimum = out.first;
		if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) out.second << '(';
		bool first = true;
		for (const auto& element : alternation.getElements()) {
			if(first) {
				first = false;
			} else {
				out.second << '+';
			}
			const UnboundedRegExpElement& object = static_cast<const UnboundedRegExpElement&>(*element);
			out.first = Priority::ALTERNATION;
			object.Accept(userData, *this);
		}
		if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) out.second << ')';
	}
}

void RegExpToStringComposer::Visit(void* userData, const UnboundedRegExpConcatenation& concatenation) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);

	Priority outerPriorityMinimum = out.first;
	if(concatenation.getElements().size() == 0) {
		out.second << "#E";
	} else if(concatenation.getElements().size() == 1) {
		const auto& element = concatenation.getElements()[0];
		const UnboundedRegExpElement& object = static_cast<const UnboundedRegExpElement&>(*element);
		object.Accept(userData, *this);
	} else {
		if(outerPriorityMinimum == Priority::FACTOR) out.second << '(';
		bool first = true;
		for (auto element : concatenation.getElements()) {
			if(first) 
				first = false;
			else
				out.second << ' ';
			const UnboundedRegExpElement& object = static_cast<const UnboundedRegExpElement&>(*element);
			out.first = Priority::CONCATENATION;
			object.Accept(userData, *this);
		}
		if(outerPriorityMinimum == Priority::FACTOR) out.second << ')';
	}
}

void RegExpToStringComposer::Visit(void* userData, const UnboundedRegExpIteration& iteration) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);

	const UnboundedRegExpElement& object = static_cast<const UnboundedRegExpElement&>(iteration.getElement());
	out.first = Priority::FACTOR;
	object.Accept(userData, *this);
	out.second << "*";
}

void RegExpToStringComposer::Visit(void* userData, const UnboundedRegExpSymbol& symbol) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);
	
	alphabet::SymbolToStringComposer composer;
	out.second << composer.compose(symbol.getSymbol());
}

void RegExpToStringComposer::Visit(void* userData, const UnboundedRegExpEpsilon&) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);
	out.second << "#E";
}

void RegExpToStringComposer::Visit(void* userData, const UnboundedRegExpEmpty&) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);
	out.second << "#0";
}

void RegExpToStringComposer::Visit(void* userData, const FormalRegExpAlternation& alternation) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);

	Priority outerPriorityMinimum = out.first;
	if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) out.second << '(';

	out.first = Priority::ALTERNATION;

	{
		const auto& element = alternation.getLeftElement();
		const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(element);
		object.Accept(userData, *this);
	}

	out.second << '+';

	{
		const auto& element = alternation.getRightElement();
		const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(element);
		object.Accept(userData, *this);
	}

	if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) out.second << ')';
}

void RegExpToStringComposer::Visit(void* userData, const FormalRegExpConcatenation& concatenation) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);

	Priority outerPriorityMinimum = out.first;
	if(outerPriorityMinimum == Priority::FACTOR) out.second << '(';

	out.first = Priority::CONCATENATION;

	{
		const auto& element = concatenation.getLeftElement();
		const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(element);
		object.Accept(userData, *this);
	}

	out.second << ' ';

	{
		const auto& element = concatenation.getRightElement();
		const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(element);
		object.Accept(userData, *this);
	}

	if(outerPriorityMinimum == Priority::FACTOR) out.second << ')';
}

void RegExpToStringComposer::Visit(void* userData, const FormalRegExpIteration& iteration) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);

	const FormalRegExpElement& object = static_cast<const FormalRegExpElement&>(iteration.getElement());
	out.first = Priority::FACTOR;
	object.Accept(userData, *this);
	out.second << "*";
}

void RegExpToStringComposer::Visit(void* userData, const FormalRegExpSymbol& symbol) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);
	
	alphabet::SymbolToStringComposer composer;
	out.second << composer.compose(symbol.getSymbol());
}

void RegExpToStringComposer::Visit(void* userData, const FormalRegExpEpsilon&) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);
	out.second << "#E";
}

void RegExpToStringComposer::Visit(void* userData, const FormalRegExpEmpty&) {
	std::pair<Priority, std::stringstream> &out = *((std::pair<Priority, std::stringstream>*) userData);
	out.second << "#0";
}

std::string RegExpToStringComposer::compose(const RegExp& regexp) {
	std::pair<Priority, std::stringstream> out;
	out.first = Priority::ALTERNATION;
	regexp.getData().Accept((void*) &out, *this);
	return std::move(out.second).str();
}

} /* namespace regexp */
