/*
 * UnboundedRegExp.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "UnboundedRegExp.h"
#include "../../exception/AlibException.h"
#include "UnboundedRegExpEmpty.h"
#include "UnboundedRegExpSymbol.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../../std/set.hpp"
#include "../formal/FormalRegExp.h"

namespace regexp {

UnboundedRegExp::UnboundedRegExp() {
	this->regExp = new UnboundedRegExpEmpty();
}

UnboundedRegExp::UnboundedRegExp(const FormalRegExp& other) : alphabet(other.getAlphabet()) {
	this->regExp = NULL;
	UnboundedRegExpElement* element = other.getRegExp().cloneAsUnbounded();
	setRegExp(std::move(*element));
	delete element;
}

UnboundedRegExp::UnboundedRegExp(const std::set<alphabet::Symbol>& alphabet, const UnboundedRegExpElement& regExp) : alphabet(alphabet) {
	this->regExp = NULL;
	setRegExp(regExp);
}

UnboundedRegExp::UnboundedRegExp(std::set<alphabet::Symbol>&& alphabet, UnboundedRegExpElement&& regExp) : alphabet(std::move(alphabet)) {
	this->regExp = NULL;
	setRegExp(std::move(regExp));
}

UnboundedRegExp::UnboundedRegExp(const UnboundedRegExpElement& regExp) {
	regExp.computeMinimalAlphabet(alphabet);
	this->regExp = NULL;
	setRegExp(regExp);
}

UnboundedRegExp::UnboundedRegExp(UnboundedRegExpElement&& regExp) {
	regExp.computeMinimalAlphabet(alphabet);
	this->regExp = NULL;
	setRegExp(std::move(regExp));
}

UnboundedRegExp::UnboundedRegExp(const UnboundedRegExp& other) : regExp(other.regExp->clone()), alphabet(other.alphabet) {
	this->regExp->attachRegExp(this);
}

UnboundedRegExp::UnboundedRegExp(UnboundedRegExp&& other) noexcept : regExp(other.regExp), alphabet(std::move(other.alphabet) ) {
	this->regExp->attachRegExp(this);
	other.regExp = NULL;
}

RegExpBase* UnboundedRegExp::clone() const {
	return new UnboundedRegExp(*this);
}

RegExpBase* UnboundedRegExp::plunder() && {
	return new UnboundedRegExp(std::move(*this));
}

UnboundedRegExp& UnboundedRegExp::operator=(const UnboundedRegExp& other) {
	if (this == &other) {
		return *this;
	}

	*this = UnboundedRegExp(other);

	return *this;
}

UnboundedRegExp& UnboundedRegExp::operator=(UnboundedRegExp&& other) noexcept {
	std::swap(this->regExp, other.regExp);
	std::swap(this->alphabet, other.alphabet);
	return *this;
}

UnboundedRegExp::~UnboundedRegExp() noexcept {
	delete regExp;
}

const UnboundedRegExpElement& UnboundedRegExp::getRegExp() const {
	return *regExp;
}

UnboundedRegExpElement& UnboundedRegExp::getRegExp() {
	return *regExp;
}

void UnboundedRegExp::setRegExp(const UnboundedRegExpElement& regExp) {
	delete this->regExp;
	this->regExp = regExp.clone();
	if(!this->regExp->attachRegExp(this))
		throw exception::AlibException("Input symbols not in the alphabet.");
}

void UnboundedRegExp::setRegExp(UnboundedRegExpElement&& regExp) {
	delete this->regExp;
	this->regExp = std::move(regExp).plunder();
	if(!this->regExp->attachRegExp(this))
		throw exception::AlibException("Input symbols not in the alphabet.");
}

const std::set<alphabet::Symbol>& UnboundedRegExp::getAlphabet() const {
	return alphabet;
}

bool UnboundedRegExp::addSymbolToAlphabet(const alphabet::Symbol & symbol) {
	return alphabet.insert(symbol).second;
}

void UnboundedRegExp::setAlphabet(const std::set<alphabet::Symbol> & symbols) {
	std::set<alphabet::Symbol> minimalAlphabet;
	this->regExp->computeMinimalAlphabet(minimalAlphabet);
	std::set<alphabet::Symbol> removedSymbols;
	std::set_difference(minimalAlphabet.begin(), minimalAlphabet.end(), symbols.begin(), symbols.end(), std::inserter(removedSymbols, removedSymbols.end()));

	if(removedSymbols.size() > 0)
		throw exception::AlibException("Input symbols are used.");

	this->alphabet = symbols;
}

bool UnboundedRegExp::removeSymbolFromAlphabet(const alphabet::Symbol & symbol) {
	if(this->regExp->testSymbol(symbol))
		throw exception::AlibException("Input symbol \"" + (std::string) symbol + "\" is used.");
	
	return alphabet.erase(symbol);
}

void UnboundedRegExp::operator >>(std::ostream& out) const {
	out << "(UnboundedRegExp " << *(this->regExp) << ")";
}

bool UnboundedRegExp::operator<(const UnboundedRegExp& other) const {
	return std::tie(*regExp, alphabet) < std::tie(*other.regExp, other.alphabet);
}

bool UnboundedRegExp::operator==(const UnboundedRegExp& other) const {
	return *regExp == *other.regExp && this->alphabet == other.alphabet;
}

bool UnboundedRegExp::operator<(const ObjectBase& other) const {
	return other > *this;
}

bool UnboundedRegExp::operator ==(const ObjectBase& other) const {
	return other == *this;
}

bool UnboundedRegExp::operator >(const ObjectBase& other) const {
	return other < *this;
}

UnboundedRegExp::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str(); 
}

} /* namespace regexp */
