/* addr2line.c -- convert addresses to line number and function name
 * Copyright (C) 1997-2014 Free Software Foundation, Inc.
 * Contributed by Ulrich Lauther <Ulrich.Lauther@mchp.siemens.de>
 *
 * This file is part of GNU Binutils.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 51 Franklin Street - Fifth Floor, Boston,
 * MA 02110-1301, USA. */

/* Derived from objdump.c and nm.c by Ulrich.Lauther@mchp.siemens.de
 * modified for purposes of stacktrace generation by
 * jan.travnicek@fit.cvut.cz */

#define PACKAGE "alib2data"
#define PACKAGE_VERSION "1"

#include "addr2line.h"
#include <bfd.h>
#include <sstream>

off_t get_file_size ( const char *file_name ) {
	struct stat statbuf;

	if ( stat ( file_name, &statbuf ) < 0 )
		return -1;
	else if ( !S_ISREG ( statbuf.st_mode ) )
		return -1;
	else if ( statbuf.st_size < 0 )
		return -1;
	else
		return statbuf.st_size;

	return (off_t)-1;
}

/* Read in the symbol table. */
asymbol** slurp_symtab ( bfd *abfd ) {
	long storage;
	long symcount;
	bfd_boolean dynamic = FALSE;

	if ( ( bfd_get_file_flags ( abfd ) & HAS_SYMS ) == 0 )
		return NULL;

	storage = bfd_get_symtab_upper_bound ( abfd );

	if ( storage == 0 ) {
		storage = bfd_get_dynamic_symtab_upper_bound ( abfd );
		dynamic = TRUE;
	}

	if ( storage < 0 )
		return NULL;

	asymbol** syms = (asymbol **) malloc ( storage * sizeof(asymbol*) );

	if ( dynamic )
		symcount = bfd_canonicalize_dynamic_symtab ( abfd, syms );
	else
		symcount = bfd_canonicalize_symtab ( abfd, syms );

	if ( symcount < 0 )
		return NULL;

	/* If there are no symbols left after canonicalization and we have not tried the dynamic symbols then give them a go. */
	if ( ( symcount == 0 ) && !dynamic && ( ( storage = bfd_get_dynamic_symtab_upper_bound ( abfd ) ) > 0 ) ) {
		free ( syms );
		syms = (asymbol**) malloc ( storage * sizeof(asymbol*) );
		symcount = bfd_canonicalize_dynamic_symtab ( abfd, syms );
	}

	return syms;
}

typedef struct {
	bfd_vma vma;
	bfd_vma pc;
	asection* section;
} bfd_find_section_data;

/* Look for a section. This is called via bfd_map_over_sections. */
void find_section ( bfd *abfd, asection *section, void* data ) {
	bfd_vma pc = ((bfd_find_section_data*) data)->pc;

	if ( ( bfd_get_section_flags ( abfd, section ) & SEC_ALLOC ) == 0 )
		return;

	bfd_vma vma = bfd_get_section_vma ( abfd, section );

	if ( pc < vma )
		return;

	bfd_size_type size = bfd_get_section_size ( section );

	if ( pc >= vma + size )
		return;

	((bfd_find_section_data*) data)->vma = vma;
	((bfd_find_section_data*) data)->section = section;
}

std::string translate_addresses ( bfd *abfd, asection *section, asymbol** syms, const char *addr_hex ) {
	bfd_vma pc = bfd_scan_vma ( addr_hex, NULL, 16 );
	bfd_vma vma = 0;
	const char * filename;
	const char * functionname;
	unsigned int line;
	unsigned int discriminator;

	std::stringstream ss;

	ss << "0x";
	//bfd_printf_vma ( abfd, pc );
	ss << std::hex << pc << std::dec;

	ss << ": ";

	bfd_boolean found = FALSE;

	if ( ! section ) {
		bfd_find_section_data data = {0, pc, NULL};
		bfd_map_over_sections ( abfd, find_section, &data );
		section = data.section;
		vma = data.vma;
	}

	if ( ( bfd_get_section_flags ( abfd, section ) & SEC_ALLOC ) == 0 ) {
		ss << "?? ";
		return std::move(ss).str();
	}
	
	if ( pc < vma ) {
		ss << "Corrupted ";
		return std::move(ss).str();
	}

	bfd_size_type size = bfd_get_section_size ( section );

	if ( pc >= vma + size ) {
		ss << "?? ";
		return std::move(ss).str();
	}

	found = bfd_find_nearest_line_discriminator ( abfd, section, syms, pc - vma, &filename, &functionname, &line, &discriminator );

	if ( !found ) {
		ss << "?? ";
		return std::move(ss).str();
	}

	while ( 1 ) {
		const char * name;
		char * alloc = NULL;

		name = functionname;

		if ( ( name == NULL ) || ( *name == '\0' ) ) {
			name = "??";
		} else {
			alloc = bfd_demangle ( abfd, name, 0 );

			if ( alloc != NULL )
				name = alloc;
		}

		ss << name;

		ss << " at ";

		if ( alloc != NULL )
			free ( alloc );

		ss << (filename ? filename : "??") << ":";

		if ( line != 0 ) {
			if ( discriminator != 0 )
				ss << line << " (discriminator " << discriminator << ")";
			else
				ss << line;
		} else {
			ss << "?";
		}

		found = bfd_find_inliner_info ( abfd, &filename, &functionname, &line );

		if ( !found )
			break;

		ss << std::endl << " (inlined by) ";
	}

	return std::move(ss).str();
}

/* Process a file. Returns an exit value for main(). */
std::string addrToLine ( const char *file_name, const char *section_name, const char *target, const char *addr_hex ) {
	bfd * abfd;
	asection *section;
	char ** matching;
	asymbol** syms;

	if ( get_file_size ( file_name ) < 1 )
		return "";

	abfd = bfd_openr ( file_name, target );

	if ( abfd == NULL )
		return "";

	/* Decompress sections. */
	abfd->flags |= BFD_DECOMPRESS;

	if ( bfd_check_format ( abfd, bfd_archive ) ) {
		std::stringstream ss;
		ss << file_name << ": cannot get addresses from archive";
		return std::move(ss).str();
	}

	if ( !bfd_check_format_matches ( abfd, bfd_object, &matching ) ) {
		if ( bfd_get_error ( ) == bfd_error_file_ambiguously_recognized )
			free ( matching );

		return "";
	}

	if ( section_name != NULL ) {
		section = bfd_get_section_by_name ( abfd, section_name );

		if ( section == NULL ) {
			printf ( "%s: cannot find section %s", file_name, section_name );
			return "";
		}
	} else {
		section = NULL;
	}

	if ( !(syms = slurp_symtab ( abfd )) )
		return "";

	std::string res = translate_addresses ( abfd, section, syms, addr_hex );

	if ( syms != NULL ) {
		free ( syms );
		syms = NULL;
	}

	bfd_close ( abfd );

	return res;
}
