#include <list>
#include "GrammarTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "grammar/UnknownGrammar.h"

#include "grammar/Unrestricted/UnrestrictedGrammar.h"
#include "grammar/GrammarFromXMLParser.h"
#include "grammar/GrammarToXMLComposer.h"

#include "factory/DataFactory.hpp"

#include "alphabet/LabeledSymbol.h"

#include "grammar/GrammarFromStringParser.h"
#include "grammar/GrammarToStringComposer.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_REGISTRATION( GrammarTest );

void GrammarTest::setUp() {
}

void GrammarTest::tearDown() {
}

void GrammarTest::testXMLParser() {
	grammar::UnknownGrammar grammar;
	grammar.setNonterminalAlphabet({alphabet::symbolFrom(1), alphabet::symbolFrom(2), alphabet::symbolFrom(3)});
	grammar.setTerminalAlphabet({alphabet::symbolFrom("a"), alphabet::symbolFrom("b")});
	grammar.addRule(grammar::UnknownRule());

	CPPUNIT_ASSERT( grammar == grammar );
	{
		std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::list<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		grammar::UnknownGrammar grammar2 = alib::DataFactory::fromTokens<grammar::UnknownGrammar>(tokens2);

		CPPUNIT_ASSERT( grammar == grammar2 );
	}
	{
		std::string tmp = alib::DataFactory::toString(grammar);
		grammar::Grammar grammar2 = alib::DataFactory::fromString<grammar::Grammar>(tmp);

		CPPUNIT_ASSERT( grammar == grammar2.getData() );
	}
}

void GrammarTest::testUnrestrictedParser() {
	grammar::UnrestrictedGrammar grammar(alphabet::symbolFrom(1));

	grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
	grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
	grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
	grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
	grammar.addTerminalSymbol(alphabet::symbolFrom("b"));

	grammar.addRule({alphabet::symbolFrom(1)}, {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
	grammar.addRule({alphabet::symbolFrom(2)}, {alphabet::symbolFrom("b"), alphabet::symbolFrom(3)});
	grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(2), alphabet::symbolFrom(2)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
	grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(2)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("b"), alphabet::symbolFrom(3)});
	grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(1)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom(3)});
	grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(1)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
	grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(2)}, std::vector<alphabet::Symbol> {});

	CPPUNIT_ASSERT( grammar == grammar );
	{
		std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::list<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		grammar::UnrestrictedGrammar grammar2 = alib::DataFactory::fromTokens<grammar::UnrestrictedGrammar>(tokens2);

		CPPUNIT_ASSERT( grammar == grammar2 );
	}
	{
		grammar::ContextPreservingUnrestrictedGrammar grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));

		grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)}, alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
		grammar.addRule(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(2), std::vector<alphabet::Symbol> {}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("b"), alphabet::symbolFrom(3)});
		grammar.addRule(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {}, std::vector<alphabet::Symbol> {alphabet::symbolFrom(3)});
		grammar.addRule(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
		grammar.addRule(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(2), std::vector<alphabet::Symbol> {}, std::vector<alphabet::Symbol> {});
		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::ContextPreservingUnrestrictedGrammar grammar2 = alib::DataFactory::fromTokens<grammar::ContextPreservingUnrestrictedGrammar>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
}

void GrammarTest::testRegularParser() {
	{
		grammar::RightRG grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));

		grammar.addRule(alphabet::symbolFrom(1), std::make_pair(alphabet::symbolFrom("a"), alphabet::symbolFrom(2)));
		grammar.addRule(alphabet::symbolFrom(2), std::make_pair(alphabet::symbolFrom("b"), alphabet::symbolFrom(3)));

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::RightRG grammar2 = alib::DataFactory::fromTokens<grammar::RightRG>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}

	{
		grammar::RightLG grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));

		grammar.addRule(alphabet::symbolFrom(1), std::make_pair(std::vector<alphabet::Symbol> {alphabet::symbolFrom("a")}, alphabet::symbolFrom(2)));
		grammar.addRule(alphabet::symbolFrom(2), std::make_pair(std::vector<alphabet::Symbol> {alphabet::symbolFrom("b")}, alphabet::symbolFrom(3)));
		grammar.addRule(alphabet::symbolFrom(1), std::make_pair(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(3)));
		grammar.addRule(alphabet::symbolFrom(1), std::make_pair(std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom("a")}, alphabet::symbolFrom(2)));

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::RightLG grammar2 = alib::DataFactory::fromTokens<grammar::RightLG>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
}

void GrammarTest::testContextFreeParser() {
	{
		grammar::LG grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));

		grammar.addRule(alphabet::symbolFrom(1), std::make_tuple(std::vector<alphabet::Symbol> {alphabet::symbolFrom("a")}, alphabet::symbolFrom(2), std::vector<alphabet::Symbol>{}));
		grammar.addRule(alphabet::symbolFrom(2), std::make_tuple(std::vector<alphabet::Symbol> {alphabet::symbolFrom("b")}, alphabet::symbolFrom(3), std::vector<alphabet::Symbol>{}));
		grammar.addRule(alphabet::symbolFrom(1), std::make_tuple(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(3), std::vector<alphabet::Symbol>{}));
		grammar.addRule(alphabet::symbolFrom(1), std::make_tuple(std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom("a")}, alphabet::symbolFrom(2), std::vector<alphabet::Symbol>{}));

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::LG grammar2 = alib::DataFactory::fromTokens<grammar::LG>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
	{
		grammar::EpsilonFreeCFG grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));
		grammar.setGeneratesEpsilon(true);

		grammar.addRule(alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
		grammar.addRule(alphabet::symbolFrom(2), std::vector<alphabet::Symbol> {alphabet::symbolFrom("b"), alphabet::symbolFrom(3)});
		grammar.addRule(alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {alphabet::symbolFrom(3)});
		grammar.addRule(alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::EpsilonFreeCFG grammar2 = alib::DataFactory::fromTokens<grammar::EpsilonFreeCFG>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
	{
		grammar::CFG grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));

		grammar.addRule(alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
		grammar.addRule(alphabet::symbolFrom(2), std::vector<alphabet::Symbol> {alphabet::symbolFrom("b"), alphabet::symbolFrom(3)});
		grammar.addRule(alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {alphabet::symbolFrom(3)});
		grammar.addRule(alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
		grammar.addRule(alphabet::symbolFrom(2), std::vector<alphabet::Symbol> {});

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::CFG grammar2 = alib::DataFactory::fromTokens<grammar::CFG>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
	{
		grammar::CNF grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));
		grammar.setGeneratesEpsilon(true);

		grammar.addRule(alphabet::symbolFrom(1), std::make_pair(alphabet::symbolFrom(3), alphabet::symbolFrom(2)));
		grammar.addRule(alphabet::symbolFrom(2), std::make_pair(alphabet::symbolFrom(3), alphabet::symbolFrom(3)));
		grammar.addRule(alphabet::symbolFrom(3), alphabet::symbolFrom("a"));
		grammar.addRule(alphabet::symbolFrom(1), alphabet::symbolFrom("a"));

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::CNF grammar2 = alib::DataFactory::fromTokens<grammar::CNF>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
	{
		grammar::GNF grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));
		grammar.setGeneratesEpsilon(true);

		grammar.addRule(alphabet::symbolFrom(1), std::make_pair(alphabet::symbolFrom("a"), std::vector<alphabet::Symbol> {alphabet::symbolFrom(2)}));
		grammar.addRule(alphabet::symbolFrom(2), std::make_pair(alphabet::symbolFrom("b"), std::vector<alphabet::Symbol> {alphabet::symbolFrom(3)}));
		grammar.addRule(alphabet::symbolFrom(3), std::make_pair(alphabet::symbolFrom("a"), std::vector<alphabet::Symbol> {}));
		grammar.addRule(alphabet::symbolFrom(1), std::make_pair(alphabet::symbolFrom("a"), std::vector<alphabet::Symbol> {}));

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::GNF grammar2 = alib::DataFactory::fromTokens<grammar::GNF>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
}

void GrammarTest::testContextSensitiveParser() {
	{
		grammar::CSG grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));

		grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)}, alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
		grammar.addRule(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(2), std::vector<alphabet::Symbol> {}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("b"), alphabet::symbolFrom(3)});
		grammar.addRule(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {}, std::vector<alphabet::Symbol> {alphabet::symbolFrom(3)});
		grammar.addRule(std::vector<alphabet::Symbol> {}, alphabet::symbolFrom(1), std::vector<alphabet::Symbol> {}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::CSG grammar2 = alib::DataFactory::fromTokens<grammar::CSG>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
	{
		grammar::NonContractingGrammar grammar(alphabet::symbolFrom(1));

		grammar.addNonterminalSymbol(alphabet::symbolFrom(1));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(2));
		grammar.addNonterminalSymbol(alphabet::symbolFrom(3));
		grammar.addTerminalSymbol(alphabet::symbolFrom("a"));
		grammar.addTerminalSymbol(alphabet::symbolFrom("b"));

		grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(2), alphabet::symbolFrom(2)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});
		grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(2)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("b"), alphabet::symbolFrom(3)});
		grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(1)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom(3)});
		grammar.addRule(std::vector<alphabet::Symbol> {alphabet::symbolFrom(1)}, std::vector<alphabet::Symbol> {alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom("a"), alphabet::symbolFrom(2)});

		CPPUNIT_ASSERT( grammar == grammar );
		{
			std::list<sax::Token> tokens = alib::DataFactory::toTokens(grammar);
			std::string tmp;
			sax::SaxComposeInterface::printMemory(tmp, tokens);

			std::list<sax::Token> tokens2;
			sax::SaxParseInterface::parseMemory(tmp, tokens2);
			grammar::NonContractingGrammar grammar2 = alib::DataFactory::fromTokens<grammar::NonContractingGrammar>(tokens2);

			CPPUNIT_ASSERT( grammar == grammar2 );
		}
	}
}

void GrammarTest::testStringParser() {
	{
	}
	{
		std::string input = 	"({0, 1, 2, 3},\n"
					" {a, b},\n"
					" {0 -> #E | a 1,\n"
					"  1 -> #E | b 2,\n"
					"  2 -> #E | a 3,\n"
					"  3 -> a | b},\n"
					" S)\n";
		std::stringstream inputs(input);

		grammar::GrammarFromStringParser parser(inputs);
		grammar::Grammar grammar = parser.parseValue();

		std::string input2 = 	"({0, 1, 2, 3},\n"
					" {a, b},\n"
					" {0 -> a 1,\n"
					"  0 -> #E,\n"
					"  1 -> #E | b 2,\n"
					"  2 -> a 3 | #E,\n"
					"  3 -> a | b | a},\n"
					" S)\n";
		std::stringstream inputs2(input2);

		grammar::GrammarFromStringParser parser2(inputs2);
		grammar::Grammar grammar2 = parser2.parseValue();

		CPPUNIT_ASSERT( grammar == grammar2 );

		grammar::GrammarToStringComposer composer;
		std::string output = composer.compose(grammar);
		std::stringstream outputs(output);

		CPPUNIT_ASSERT( input == output );

		grammar::GrammarFromStringParser parser3(outputs);
		grammar::Grammar grammar3 = parser3.parseValue();

		CPPUNIT_ASSERT( grammar == grammar3 );
	}
}
